/***************************************************************************/
/***                                                                     ***/
/***                  gencsp.c                                           ***/
/***                                                                     ***/
/***       Ronny Fehling, Bernhard Nebel, Jochen Renz  - March 1998      ***/
/***                                                                     ***/
/***         fehling, nebel, renz@informatik.uni-freiburg.de             ***/
/***                                                                     ***/
/***          http://www.informatik.uni-freiburg.de/~sppraum             ***/
/***                                                                     ***/
/***                      Institut fuer Informatik                       ***/
/***                     Albert-Ludwigs-Universitaet                     ***/
/***                           Am Flughafen 17                           ***/
/***                       79110 Freiburg, Germany                       ***/
/***                                                                     ***/
/***************************************************************************/


/*
    gencsp [-au] [-is number] [-cdln min [max step]]  [-rx file] [-o name]

    Generates RCC8 CSPs. 
    options are:
       -a   adds relations to labels to make the CSP consistent (van Beek)
       -u   means to use an uniform distribution (excluding CSP with zero labs)

       -i <number>  specifies number of instances to be generated (default 1)
       -s <number>  specifies the seed for the random function

      The following options expect at least one value or three indicating 
      minimum value, maximum value and the stepwidth.
      Numbers in brackets specify the range of accepted values.
       -c   specifies the constraint density (default 0.0) [0.01 - 1.0]
       -d   specifies the avgerage degree of a node (default 0) [0 - MAXCSP]
       -l   specifies the labelsize (default 4.0) [1.0 - 8.0] 
       -n   specifies the number of nodes (default 3) [3 - MAXCSP]

       -r <file>  restricts all relations to be the ones mentioned in the file
       -x <file>  specifies the relations to be excluded

       -o <name>  specifies the prefix of the outputfile 
                  (default is none)

       The -c and -d options override each other and are overriden by -u.
*/


const char *usagestring = "\
usage: gencsp [-au] [-is number] [-cdln min [max step]]  [-rx file] [-o name]

    Generates RCC8 CSPs. 
    options are:
       -a   adds relations to labels to make the CSP consistent (van Beek)
       -u   means to use an uniform distribution (excluding CSP with zero labs)

       -i <number>  specifies number of instances to be generated (default 1)
       -s <number>  specifies the seed for the random function

      The following options expect at least one value or three indicating 
      minimum value, maximum value and the stepwidth.
      Numbers in brackets specify the range of accepted values.
       -c   specifies the constraint density (default 0.0) [0.01 - 1.0]
       -d   specifies the avgerage degree of a node (default 0) [0 - MAXCSP]
       -l   specifies the labelsize (default 4.0) [1.0 - 8.0] 
       -n   specifies the number of nodes (default 3) [3 - MAXCSP]

       -r <file>  restricts all relations to be the ones mentioned in the file
       -x <file>  specifies the relations to be excluded

       -o <name>  specifies the prefix of the outputfile 
                  (default is none)

       The -c and -d options override each other and are overriden by -u.";



#include "rcc8.h"
#include "rcc8io.h"
#include "rcc8op.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/resource.h>

/* should be in stdlib.h ! */
void srand48(long seedval);
long lrand48(void);

/* switches for gencsp */ 
unsigned int   swseed = 0;
int swaddcons = 0;
int swoutfile = 0;
int swinstances = 1;
int swminnodes  = 3;
int swmaxnodes = 3;
int swstepnodes = 1;
int swexclude = 0;
int swrestrict = 0;
int restrictsize;
int swuniform = 0;
float swmindegree = 0.0;
float swmaxdegree = 0.0;
float swstepdegree = 1.0;
float swminlabsiz = 4.0;
float swmaxlabsiz = 4.0;
float swsteplabsiz = 1.0;
float swminconsdensity = 0.0;
float swmaxconsdensity = 0.0;
float swstepconsdensity = 1.0;


/* statistical info */
char cstype[100]; 
char csptype[100];
char xalg[MAXSET];
char ralg[MAXSET];
RELTYPE ralglist[MAXSET];
RELTYPE splitinfo[MAXSET][9];

#if defined(DYNAMIC)
RELTYPE **csp;
#else 
RELTYPE csp[MAXCSP][MAXCSP];
#endif


/* files and filenames */
FILE *infile;
char infilename[200];
FILE *splitfile;
char splitfilename[200];
FILE *outfile;
char outfilename[200];
char ralgname[200];
char xalgname[200];


/* seed for random variable */
unsigned int randomseed = 0;



/* Initalizes random seed */
void initrand(unsigned int seed)
{
  if (seed == 0)
    randomseed = time(NULL);
  else
    randomseed = seed;

  srand48(randomseed);
}



/* Makes a CSP consistent */
void makeconsistent(csp, nodes)
     int nodes;
#if defined(DYNAMIC)
     RELTYPE   **csp;
#else 
     RELTYPE csp[MAXCSP][MAXCSP];
#endif
{
  register int i,j;

  int e[MAXCSP], s[MAXCSP];

  /* adopted code from van Beek */
  for (i = 0; i < nodes; i++) {
    s[i] = ((abs(lrand48()))%71)+1;
    e[i] = s[i] + ((abs(lrand48()))%17)+1;
  }

  for (i = 0; i <= nodes; i++)
    for (j = i+1; j <= nodes; j++)
      if      (i == j)			        csp [i][j] |= EQ;
      else if ((e[i] <  s[j])||(e[j] <  s[i]))	csp [i][j] |= DC ;
      else if ((e[i] == s[j])||(e[j] == s[i]))	csp [i][j] |= EC ;
      else if (s[i] >  s[j] && e[i] <  e[j])	csp [i][j] |= NTPP ;
      else if (s[i] <  s[j] && e[i] >  e[j])	csp [i][j] |= NTPPI;
      else if (s[i] == s[j] && e[i] <  e[j])	csp [i][j] |= TPP ;
      else if (s[i] == s[j] && e[i] >  e[j])	csp [i][j] |= TPPI;
      else if (s[i] >  s[j] && e[i] == e[j])	csp [i][j] |= TPP ;
      else if (s[i] <  s[j] && e[i] == e[j])	csp [i][j] |= TPPI;
      else if (s[i] == s[j] && e[i] == e[j])	csp [i][j] |= EQ;
      else if (s[i] <  s[j] && e[i] <  e[j]
	       && e[i] >  s[j])	csp [i][j] |= PO ;
      else if (s[i] >  s[j] && e[i] >  e[j]
	       && s[i] <  e[j])	csp [i][j] |= PO;
      else {
	fprintf (stderr,"error:s[%d] = %d, e[%d] = %d\n",i,s[i],i,e[i]);
	fprintf (stderr,"      s[%d] = %d, e[%d] = %d\n",j,s[j],j,e[j]);
      }
}



RELTYPE genlab(float avglabsiz, int uniform, int restrict, int exclude)
{
  long int thres;
  RELTYPE rel, shift;
  int onerelx;
  int i;
  int disjcnt;
  
  if (uniform) {
    thres = 4999;
    onerelx = -1;
  } 
  else {
    thres = ((int)((avglabsiz-1.0)*10000.0/7.0))-1;
    onerelx = abs(lrand48())%8;
  }
 
  do {
    shift = 1;
    i = 0;
    rel = 0;
    disjcnt = 1;

    do {
      if (i == onerelx) {
	rel |= shift;
	disjcnt++;
      }
      else
	if (abs(lrand48())%10000 <= thres) {
	  rel |= shift;
	  disjcnt++;
	}
      shift = shift << 1;
    } while (++i < 8);
  } while ((swexclude && xalg[rel] && (disjcnt > 1)) || 
	   (swrestrict && !ralg[rel]));
  return(rel);
}



void gennet(csp,typestring,instcnt,cspsize, seed,avglabsiz,avgcondense,avgdegree)
     char *typestring;
     int instcnt,cspsize; 
     unsigned int seed;
     float avglabsiz,avgcondense,avgdegree;
#if defined(DYNAMIC)
     RELTYPE   **csp;
#else 
     RELTYPE csp[MAXCSP][MAXCSP];
#endif
{
  register int i,j;
  long int thres;
  int incons;

  
  sprintf(typestring,"#%1d-N%1d-%c%1.2f-%c%1.2f-R:%s-X:%s-S%1u",
	  instcnt,cspsize,	 
	  (avgcondense != 0.0 ? 'C' : 'D'),
	  (avgcondense != 0.0 ? avgcondense : avgdegree),
	  (swuniform ? 'U' : 'L'),
	  avglabsiz,
	  (swrestrict ? ralgname : "-"), 
	  (swexclude ? xalgname : "-"), 
	  seed);
 
  
  if (avgcondense == 0.0) 
    avgcondense = (((avgdegree))/((float)(cspsize-1)));

  thres = ((int)(avgcondense*10000.0))-1;

  do {
    incons = 0;
    for (i = 0; i < cspsize; i++)
      for (j = i+1; j < cspsize; j++)
	if (abs(lrand48())%10000 <= thres) {
	  if (!(csp[i][j] = genlab(avglabsiz,swuniform,swrestrict,swexclude))) {
	    i = cspsize;
	    j = cspsize;
	    incons = 1;
	    break;
	  }
	} else
	  csp[i][j] = DALL;
  } while (incons);
}



void usage()
{
  fprintf(stderr,"%s\n",usagestring);
  exit(-1);
}



int int_compare(long int *i1, long int *i2)
{
  return(((*i1 < *i2) ? -1 
	  : ((*i1 == *i2) ? 0 : 1)));
}



void readsplitfile(char *s)
{
  int i,j,inp;

  if ((splitfile = fopen(s,"r")) == NULL) {
    fprintf(stderr, "***Relation-cover file '%s' not found\n",s);
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "***Format error in relation-cover file '%s'\n",s);
	exit(-1);
      } else {
	splitinfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "***No eof in relation-cover file '%s'\n",s);
      	exit(-1);
      }
}



unsigned int readintarg(int *xarg, char **argv)
{
  unsigned int result;

  (*xarg)++;
  if (sscanf(argv[*xarg],"%d",&result) != 1) usage();
  return(result);
}



float readfloatarg(int *xarg, char **argv)
{
  float result;
  
  (*xarg)++;
  if (sscanf(argv[*xarg],"%f",&result) != 1) usage();
  return(result);
}



void main(int argc, char **argv)
{

  int xarg;
  register int i,nd;
  register float cd,lbs,dg;
  FILE *algfile;
  int maxsize=4;
  char prefix[100];

  swoutfile=0;
  xarg = 1;
  outfile=stdout;   /* default is stdout */

  while (xarg < argc) {
    if (argv[xarg][0] != '-') usage();
    switch(argv[xarg][1]) {
    case 'a':
      swaddcons = 1;
      break;
    case 'c':
      if(xarg+1 >= argc) usage();
      swminconsdensity = readfloatarg(&xarg,argv);
      if((xarg+2 < argc) && (argv[xarg+1][0]!='-')){
	swmaxconsdensity = readfloatarg(&xarg,argv);
	swstepconsdensity = readfloatarg(&xarg,argv);
      }
      else{
	swmaxconsdensity = swminconsdensity;
	swstepconsdensity = 1.0;
      }
      if (swminconsdensity != 0.0) swmindegree = 0.0;
      break;
    case 'd':
      if(xarg+1 >= argc) usage();
      swmindegree = readfloatarg(&xarg,argv);
      if((xarg+2 < argc) && (argv[xarg+1][0]!='-')){
	swmaxdegree = readfloatarg(&xarg,argv);
	swstepdegree = readfloatarg(&xarg,argv);
      }
      else{
	swmaxdegree = swmindegree;
	swstepdegree = 1.0;
      }
      if (swmindegree != 0.0) swminconsdensity = 0.0;
      break;
    case 'i':
      if(xarg+1 >= argc) usage();
      swinstances = readintarg(&xarg,argv);
      break;
    case 'l':
      if(xarg+1 >= argc) usage();
      swminlabsiz = readfloatarg(&xarg,argv);
      if((xarg+2 < argc) && (argv[xarg+1][0]!='-')){
	swmaxlabsiz = readfloatarg(&xarg,argv);
	swsteplabsiz = readfloatarg(&xarg,argv);
      }
      else{
	swmaxlabsiz = swminlabsiz;
	swsteplabsiz = 1.0;
      }
      break;
    case 'n':
      if(xarg+1 >= argc) usage();
      swminnodes = readintarg(&xarg,argv);
      if((xarg+2 < argc) && (argv[xarg+1][0]!='-')){
	swmaxnodes = readintarg(&xarg,argv);
	swstepnodes = readintarg(&xarg,argv);
      }
      else{
	swmaxnodes=swminnodes;
	swstepnodes=1;
      }
      maxsize=swmaxnodes+1;
      break;
    case 'o':
      if(xarg+1 >= argc) usage();
      swoutfile=1;
      outfile=NULL;
      strcpy(prefix,argv[++xarg]);
      break;
    case 'r':
      if(xarg+1 >= argc) usage();
      swrestrict = 1;
      strcpy(ralgname,argv[++xarg]);
      if ((algfile = fopen(ralgname,"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",argv[xarg]);
	exit(-1);
      }
      if (!(parsealg(algfile,ralg))) {
	fprintf(stderr,"\n*** Error in reading rel set from '%s'\n",argv[xarg]);
	exit(-1);
      }
      fclose(algfile);
      break;
    case 's':
      if(xarg+1 >= argc) usage();
      swseed = readintarg(&xarg,argv);
      break;
    case 'u':
      swuniform = 1;
      break;
    case 'x':
      if(xarg+1 >= argc) usage();
      swexclude = 1;
      strcpy(xalgname,argv[++xarg]);
      if ((algfile = fopen(xalgname,"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",argv[xarg]);
	exit(-1);
      }
      if (!(parsealg(algfile,xalg))) { 
	fprintf(stderr,"\n*** Error in reading rel set from '%s'\n",argv[xarg]);
	exit(-1);
      }
      fclose(algfile);
      break;
    default:
      usage();
    }
    xarg++;
  }

  if ((swinstances < 1) || (swminnodes < 3) || 
      (swmaxnodes > MAXCSP) || (swminnodes > swmaxnodes) || 
      (swminlabsiz > swmaxlabsiz) || (swmindegree > swmaxdegree) ||
      (swminconsdensity > swmaxconsdensity) ||
      (swmaxdegree > MAXCSP) || (swmaxlabsiz > 8.0) || 
      (swmaxconsdensity > 1.0) ||
      ((swminlabsiz < 1.0) && !swrestrict) || 
      ((swminconsdensity < 0.009) && (swmindegree < 1.0) && !swuniform)){
    usage();
  }


  initrand(swseed);

  restrictsize = 0;
  if (swrestrict) {
    for (i=1; i<MAXSET; i++) 
      if (ralg[i] && 
	  ((i != DALL) || (swminconsdensity == 1.0)) && 
	   (!swexclude || !xalg[i]))
	ralglist[restrictsize++] = i;
    if (restrictsize == 0) {
      fprintf(stderr,"\n*** Restrict-set has size 0!\n");
      exit(-1);
    }
  }

#if defined(DYNAMIC)
  csp = (RELTYPE **)malloc(maxsize*sizeof(RELTYPE *));
  csp[0] = (RELTYPE *)malloc(maxsize*maxsize*sizeof(RELTYPE));
  for(i=1;i<maxsize;i++) {
    csp[i] = csp[0] + i * maxsize; 
  }
#endif

  for(nd=swminnodes;nd<=swmaxnodes;nd+=swstepnodes){
    for(cd=swminconsdensity;cd<=swmaxconsdensity;cd+=swstepconsdensity){
      for(dg=swmindegree;dg<=swmaxdegree;dg+=swstepdegree){
	for(lbs=swminlabsiz;lbs<=swmaxlabsiz;lbs+=swsteplabsiz){
	  if(swoutfile){
	    sprintf(outfilename,"%sN%dC%1.0fD%1.0fL%1.0f",prefix,nd,cd*100.0,dg*10.0,lbs*10.0);
	    if ((outfile = fopen(outfilename,"a")) == NULL) {
	      fprintf(stderr,"\n*** Cannot open '%s' for writing\n",outfilename);
	      exit(-1); 
	    }
	  }
	  for (i = 1; i <= swinstances; i++) {
	    gennet(csp, csptype, i, nd, randomseed,lbs, cd, dg); 
	    
	    if (swaddcons) makeconsistent(csp, nd);
	    printcsp(outfile, nd-1, csptype, csp);
	  }
	  if(swoutfile)	  fclose(outfile);
	}
      }
    }
  }
  exit(0);
}




